import os, requests
import typing
from functools import reduce
from IPython.display import Markdown, display
import sqlite3
from tabulate import tabulate
from Levenshtein import ratio


class Util:
    """
    Utilities
    """

    # https://stackoverflow.com/a/16696317/3151250
    @staticmethod
    def download_file(url, folder):
        local_filename = url.split("/")[-1]
        full_path = os.path.join(folder, local_filename)
        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            with open(full_path, "wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
        print(f"Lade herunter und speichere unter {full_path}")
        return local_filename

    @staticmethod
    def evaluate_sql(path_to_db, query, display_table=True):
        if not os.path.exists(path_to_db):
            raise FileNotFoundError()
        conn = sqlite3.connect(path_to_db)
        cursor = conn.execute(query)
        attributes = [description[0] for description in cursor.description]
        tuples = [list(x) for x in cursor.fetchall()]
        table = tabulate(tuples, attributes, tablefmt="github")
        if display_table:
            display(Markdown(table))
        conn.close()
        return attributes, tuples

    @staticmethod
    def transform_pgsql_resultset_to_list_of_lists(result):
        result_list = []
        for row_dicts in result.dicts():
            row_list = []
            for attribute, value in row_dicts.items():
                row_list.append(str(value))
            result_list.append(row_list)
        attributes = result.field_names
        return attributes, result_list

    # check sql solution
    @staticmethod
    def check_sql_solution(
        query_text,
        student_dict,
        solution,
        no_of_points,
        partial_score_exact,
        partial_score_keywords,
        partial_score_points,
    ):
        score = 0
        try:
            solution = {Util.str_sanitize(k): v for k, v in solution.items()}
            student_dict = {Util.str_sanitize(k): v for k, v in student_dict.items()}
            # check for exact match
            for x in solution:
                print("Überprüfe Spalte: ", x, " | ", solution[x])
                for y in range(len(solution[x])):
                    print("   Überprüfe Zelle:", y, " | ", solution[x][y])
                    check = Util.levenshtein_str_callback(
                        solution[x][y], student_dict[x][y]
                    )
                    if 0.8 <= check:
                        score += partial_score_exact
                    print("    > ", check, ", score: ", round(score, 4))
        except Exception as e:
            print("Exception festgestellt")
            print(f"  {type(e).__name__} {e} ")
            if type(e).__name__ == "KeyError":
                print(
                    "  ... diese Errormeldung bedeutet wahrscheinlich, dass eine Spalte fehlerhaft benannt wurde oder nicht gefunden wurde."
                )
            pass  # catch, return
        else:
            # if the solution isn't exact, apply other rules
            # = check for keywords in SQL.
            if round(score, 2) < no_of_points:
                score = 0
                print("Teilpunkte")
                for i in partial_score_keywords:
                    if i in query_text.upper():
                        score += partial_score_points
        return score

    # count solution cells in result array
    @staticmethod
    def get_solution_count(solution):
        count = 0
        for x in solution:
            for y in range(len(solution[x])):
                count += 1
        if count == 0:
            count = 1
        return count

    ###############################################################################################################
    # CHECK TABLE & HELPER FUNCTIONS
    ###############################################################################################################

    @staticmethod
    def custom_list_ratio(a: typing.List[str], b: typing.List[str]) -> int:
        # empty lists have the ratio of 1
        if len(a) == len(b) == 0:
            return 1
        h = [ratio(i, j) for (i, j) in zip(a, b)]
        n = len(h)
        return reduce(lambda x, y: x + (y / n), h, 0)

    # Mapping: Index of new list -> Index of old list
    @staticmethod
    def permute_list(list, mapping):
        """
        Returns a list which is permutated according to the list mapping.
        Mapping needs to be a list of length len(list), which contains the new order of indices of the list l
        """
        l = list.copy()
        for i in range(0, len(list)):
            list[i] = l[mapping[i]]

    # Returns mapping as list or None
    @staticmethod
    def get_permutation(
        list_student: typing.List[typing.Any],
        list_solution: typing.List[typing.Any],
        elem_name: str,
        levenshtein_callback,
        levenshtein_threshold: int,
        quiet: bool,
    ) -> typing.Optional[typing.List[str]]:
        """
        Find approximate matches between elements in the student's list and the solution list using Levenshtein distance.

        Parameters:
            - list_student (list): The list representing the student's solution.
            - list_solution (list): The list representing the correct solution.
            - elem_name (str): A string describing the type of elements in the list, used for output only
            - levenshtein_callback (callable): A function that calculates the Levenshtein distance or a similar metric between two strings.
            - levenshtein_threshold (float): The numerical threshold for considering a match valid.
            - quiet (bool): If True, suppresses print statements; if False, prints information about matching and mismatching elements.

        Returns:
            - list or None: A list length len(list_student), containing for each element the index of the best match in list_solution
                            Returns None if there is an element in the list_student having no match partner in the list_solution.

        The method compares each element in list_student to elements in list_solution using levenshtein_callback.
        Mismatches and missing elements are reported, and the method returns None if any errors are detected.

        """
        res = len(list_student) * []
        a = list_student.copy()
        b: typing.List[typing.Optional[str]] = list_solution.copy()
        error = False
        for i in range(0, len(a)):
            best = -1
            best_ratio = -1
            for j in range(0, len(b)):
                if b[j] is None:  # Skip elements that were already matched
                    continue
                new_ratio = levenshtein_callback(a[i], b[j])
                if new_ratio >= best_ratio:
                    best_ratio = new_ratio
                    best = j
            if best_ratio >= levenshtein_threshold:
                res.insert(i, best)
                if best_ratio < 1 and not quiet:
                    print(
                        f"Gefunden wurde {elem_name} '{a[i]}', verwendet wird '{b[best]}'?"
                    )
                b[best] = None  # Mark as matched
            else:
                if not quiet:
                    print(f"Fehlerhaft: {elem_name} '{a[i]}'")
                error = True

        # Check if all items have been used, if not, it was missing in student's solution
        for j in range(0, len(b)):
            if b[j] != None:
                if not quiet:
                    print(f"Fehlend: {elem_name} '{b[j]}'")
                error = True
        if error:
            return None
        return res

    @staticmethod
    def str_sanitize(s: str) -> str:
        """Sanitizes the string. Replaces umlaute and makes it lowercase"""
        return (
            str(s)
            .lower()
            .strip()
            .replace("ä", "ae")
            .replace("ö", "oe")
            .replace("ü", "ue")
        )

    @staticmethod
    def levenshtein_str_callback(a, b):
        r = ratio(Util.str_sanitize(a), Util.str_sanitize(b))
        return r

    @staticmethod
    def levenshtein_list_callback(a, b):
        return Util.custom_list_ratio(
            list(map(Util.str_sanitize, list(a))), list(map(Util.str_sanitize, list(b)))
        )

    @staticmethod
    def check_table(
        attributes_solution: typing.List[str],
        tuples_solution: typing.List[typing.List[str]],
        attributes_student: typing.List[str],
        tuples_student: typing.List[typing.List[str]],
        levenshtein_threshold: int = 1,
        quiet: bool = False,
    ):
        """
        Fuzzy compares two 'tables' considering the given levenshtein threshold.
        Parameters:
            - attributes_solutions: The attributes/column names of the correct table
            - tuples solution: The tuples of the correct table
            - attributes_student: The attributes of the student table
            - tuples_student: THe tuples of the student table
            - Levenshtein_threshold: The threshold used for comparing
            - quiet: Whether to produces output regarding matches
        """
        # Get permutation according to attributes
        attribute_mapping = Util.get_permutation(
            attributes_student,
            attributes_solution,
            "attribute",
            Util.levenshtein_str_callback,
            levenshtein_threshold,
            quiet,
        )
        if attribute_mapping == None:
            return False

        # Permute solution tuples with found mapping
        # Student's solutions are not changed to explain wrong answers
        Util.permute_list(attributes_solution, attribute_mapping)
        for tuple in tuples_solution:
            Util.permute_list(tuple, attribute_mapping)

        # Check if permutation on tuples can be found
        # If not, some tuples are wrong or missing
        tuple_mapping = Util.get_permutation(
            tuples_student,
            tuples_solution,
            "tuple",
            Util.levenshtein_list_callback,
            levenshtein_threshold,
            quiet,
        )
        if tuple_mapping == None:
            return False
        return True



"""
Created on 2022-07-11

"""
import math


class Scorer:
    """
    general scorer
    """

    def __init__(self, verbose: bool = True, debug: bool = False):
        """
        constructor

        Args:
            verbose(bool): if True show scoring checks and problems, default: True
            debug(bool): if True show debugging information, default: False
        """
        self.maxscore = 0.0  # maximum score
        self.score = 0  # total score reached
        self.verbose = verbose
        self.debug = debug

    def feedback(self, msg: str, positive: bool):
        """
        give positive or negative feedback with the given message

        Args:
            msg(str): the message to print
            positive(bool): if True add a ✅ marker else ❌
        """
        marker = "✅" if positive else "❌"
        if self.verbose:
            print(f"{msg} {marker}")

    def addScore(
        self,
        amount: float,
        check: str,
        problem: str = None,
        negativeOnProblem: bool = False,
    ):
        """
        add the given amount to self.score if there is no problem
        increment self.maxscore by 1

        Args:
            amount(float): the amount to add
            check(str): message text about what was checked
            problem(str): message text about a problem - if this is not none the score will not be increased
            negativeOnProblem(bool): if True subtract the amount if a problem occured
        """
        self.maxscore += amount
        if problem is not None:
            if negativeOnProblem:
                msg = "Ziehe ab: {:.2f}".format(amount)
                self.score -= amount
            else:
                msg = f"Keine Punkte"
            self.feedback(f"{msg} für {check} wegen: {problem}", positive=False)
            return
        else:
            self.score += amount
            self.feedback(
                "Füge {:.2f} Punkte hinzu für: {:s}".format(amount, check),
                positive=True,
            )


class MultipleChoiceScorer(Scorer):
    """
    scorer for multiple choice questions

    Für jede richtige Antwort gibt es 1/𝑛 Punkte.
    Für jede falsche Antwort gibt es −1/𝑛 Punkte.
    Es kann insgesamt nicht weniger als 0 Punkte geben. Wenn Sie also mehr falsche als richtige Antworten auswählen, gibt es insgesamt keine Minuspunkte.
    Falls zu viele Antworten angegeben wurden, gibt es 0 Punkte - auch wenn mehr richtige als falsche Antworten ausgewählt wurden. Achten Sie also darauf, nicht zu viele Antworten einzutragen.
    """

    def evaluate_mctask(self, result, correct, max_points, toManyEquals0: bool):
        """
        evaluate the given multiple choice task results
        """
        amount = max_points / len(correct)
        for ans in result:
            problem = None
            check = f"Überprüfung Antwort {ans}"
            if not ans in correct:
                problem = f" Ist falsch"
            self.addScore(amount, check, problem, negativeOnProblem=True)
        # Bei zu vielen angegebenen Antworten mehr falschen als richtigen Antworten: 0 Punkte
        if self.score < 0 or (len(result) > len(correct) and toManyEquals0 == True):
            self.score = 0
        return self.score


class SetScorer(Scorer):
    """
     scorer für Sets

    Für jede richtige Antwort gibt es 1/𝑛 Punkte.
    Für jede falsche Antwort gibt es −1/𝑛 Punkte.

    Falls zu viele Antworten angegeben wurden, gibt es negative Punkte
    Achten Sie also darauf, nicht zu viele Antworten einzutragen.

    Es kann insgesamt nicht weniger als 0 Punkte geben.
    Wenn Sie also mehr falsche als richtige Antworten auswählen, gibt es insgesamt keine Minuspunkte.
    """

    def evaluate_set(self, result, expected, max_points):
        amount = max_points / len(expected)
        for ans in result:
            problem = None
            check = f"Überprüfung {ans}"
            if not ans in expected:
                problem = f"Ist fehlerhaft"
            self.addScore(amount, check, problem, negativeOnProblem=True)
        # Bei zu vielen angegebenen Antworten mehr falschen als richtigen Antworten: 0 Punkte
        if self.score < 0:
            self.score = 0
        return self.score



class Exercise:
    """
    This class manages the state of the exercise.
    """

    def __init__(self, exc):
        """
        constructor

        Args:
            exc(int): exercise (ÜB), example: 1
            task(str): Task (Aufgabe), example: 2
            subtask(str): Subtask(Teilaufgabe), example: a
        """

        self.exc = exc
        self.tasks = {}
        pass

    def addTask(self, task):
        """
        add the given task

        Args:
            task: the task to add
        """
        if task.task in self.tasks:
            raise Exception("task already exists")
        if task.task not in self.tasks:
            self.tasks[task.task] = {}
        self.tasks[task.task] = task

    def getTasks(self):
        return self.tasks

    def getTaskByLabel(self, label):
        return self.tasks[label]

    def getTaskNo(self):
        return len(self.tasks)

    def getPoints(self):
        x = 0
        for t in self.tasks.values():
            x += t.getPoints()
        return x

    def __str__(self) -> str:
        return (
            f"Übung {self.exc}, {self.getTaskNo()} Aufgaben, {self.getPoints()} Punkte"
        )


class Task:
    def __init__(self, exc, task, points=0):
        """
        This class manages the state of the task.

        Args:
            exc(int): exercise (ÜB), example: 1
            task(str): Task (Aufgabe), example: 2
            points(float): Points (Punkte), example: 2
        """
        self.exc = exc
        self.task = task
        self.points = points
        self.subtasks = {}

        exc.addTask(self)

    def __str__(self):
        return self.getTaskInfo()

    def getTaskNo(self):
        return len(self.subtasks)

    def getTaskInfo(self):
        return f"Aufgabe {self.task}, {self.getPoints()} Punkte, {self.getTaskNo()} Teilaufgaben"

    def getPoints(self):
        x = 0
        for t in self.subtasks.values():
            x += t.getPoints()
        return x

    def getScore(self):
        x = 0
        for t in self.subtasks.values():
            x += t.getScore()
        return x

    def setSolution(self, solution):
        self.solution = solution

    def getSolution(self):
        return self.solution

    def addSubtask(self, subtask):
        self.subtasks[subtask.subtask] = subtask

    def getSubtasks(self):
        return self.subtasks

    def getSubtaskByLabel(self, label):
        return self.subtasks[label]


class SubTask:
    def __init__(self, task, subtask, points):
        """
        This class manages the state of the subtask.

        Args:
            exc(int): exercise (ÜB), example: 1
            subtask(str): Subtask (Teilaufgabe), example: 1.2
            points(float): Points (Punkte), example: 2
        """
        self.task = task
        self.subtask = subtask
        self.points = points
        self.score = 0

        task.addSubtask(self)

    def __str__(self):
        return self.getTaskInfo()

    def getTaskInfo(self):
        return f"Teilaufgabe {self.subtask}, {self.points} Punkte"

    def getPoints(self):
        return self.points

    def setScore(self, score):
        self.score = score

    def getScore(self):
        return self.score

    def setSolution(self, solution):
        self.solution = solution

    def getSolution(self):
        return self.solution

    def getData(self):
        return self.data

    def setData(self, data):
        self.data = data


# Spätere Features:
#   def getScaledPoints(self):
#   def autograde(self):



import json
from tests.basetest import Basetest
from excmanager.Task import *


class Testexercises(Basetest):
    """
    Test Excercise Manager
    """

    def setUp(self, debug=False, profile=True):
        return super().setUp(debug, profile)

    def testExc(self):
        """
        test exercise and task structure setup
        """
        exc = Exercise(1)
        task1 = Task(exc, "1.1")
        subtask1 = SubTask(task1, "a", points=2)

        solution = json.dumps({"test": 2})

        subtask1.setSolution(solution)

        debug = self.debug
        # debug=True
        if debug:
            print(exc)
            print(task1)
        if debug:
            print(exc)
            print(exc.getTasks())
            print(exc.getTaskByLabel("1.1").getSubtasks())
            print(exc.getTaskByLabel("1.1").getSubtaskbyLabel("a").getSolution())

        self.assertEqual(1, exc.getTaskNo())
        self.assertEqual(1, task1.getTaskNo())
        self.assertEqual(2, exc.getPoints())

        self.assertEqual(1, len(exc.getTaskByLabel("1.1").getSubtasks().keys()))
        self.assertEqual(
            2, exc.getTaskByLabel("1.1").getSubtaskByLabel("a").getPoints()
        )



from tests.basetest import Basetest
from excmanager.Util import Util


class TestUtil(Basetest):
    """
    test the Util module
    """

    def setUp(self, debug=False, profile=True):
        return super().setUp(debug, profile)

    def testCheckTable(self):
        """
        test checking tables
        """
        assert Util.check_table(["A"], [[]], ["A"], [[]], 0.5, True) == True

        assert Util.check_table(["A"], [[]], ["B"], [[]], 0.5, True) == False

        assert (
            Util.check_table(
                ["A.Titel", "B.Titel", "A.Genre"],
                [
                    ["Pulp Fiction", "Cloud Atlas", "Independent"],
                    ["Cloud Atlas", "Pulp Fiction", "Independent"],
                    ["Alien", "Matrix", "Sci-Fi"],
                    ["Blade Runner", "Matrix", "Sci-Fi"],
                ],
                ["A.Titel", "B.Titel", "A.Genre"],
                [
                    ["Alien", "Matrix", "Sci-Fi"],
                    ["Pulp Fiction", "Cloud Atlas", "Independant"],
                    ["Cloud Atlas", "Pulp Fiction", "Independant"],
                    ["Blade Runner", "Matrix", "Sci-Fi"],
                ],
                1,
                True,
            )
            == False
        )

        assert (
            Util.check_table(
                ["A.Titel", "B.Titel", "A.Genre"],
                [
                    ["Alien", "Matrix", "Sci-Fi"],
                    ["Pulp Fiction", "Cloud Atlas", "Independent"],
                    ["Cloud Atlas", "Pulp Fiction", "Independent"],
                    ["Blade Runner", "Matrix", "Sci-Fi"],
                ],
                ["A.Titel", "B.Titel", "A.Genre"],
                [
                    ["Pulp Fiction", "Cloud Atlas", "Independant"],
                    ["Cloud Atlas", "Pulp Fiction", "Independant"],
                    ["Alien", "Matrix", "Sci-Fi"],
                    ["Blade Runner", "Matrix", "Sci-Fi"],
                ],
                0.7,
                True,
            )
            == True
        )

        assert (
            Util.check_table(
                ["Nationalität"],
                [["FRANZÖSISCH"], ["ÖSTERREICHISCH"]],
                ["  nationalitaet     "],
                [["oesterreichisch"], ["     fransoesischh  "]],
                0.8,
                False,
            )
            == True
        )

        # Complete test of:
        # * umlauts
        # * spelling errors
        # * permutation of attributes
        # * permutation of tuples
        # * whitespace
        # * ints instead of strings
        # * tuples instead of lists
        assert (
            Util.check_table(
                ["Nationalität", "Alter"],
                [["FRANZÖSISCH", "2"], ["ÖSTERREICHISCH", "1"]],
                ["alterr", "  nationalitaet     "],
                [(1, "oesterreichisch"), (2, "     fransoesisch  ")],
                0.7,
                False,
            )
            == True
        )

        assert (
            Util.check_table(
                ["Vorname", "Nachname"],
                [
                    ["Quentin", "Tarantino"],
                    ["Robert", "Rodriguez"],
                    ["Ridley", "Scott"],
                    ["Lana", "Wachowski"],
                ],
                ["Vornname", "Nachname"],
                [
                    ["Quendin", "Tarantino"],
                    ["Robert", "Rodriguez"],
                    ["Ridley", "Scott"],
                    ["Lana", "Wachowski"],
                ],
                0.8,
                False,
            )
            == True
        )



"""
Created on 2022-07-11

@author: wf
"""
from tests.basetest import Basetest
from excmanager.scorer import Scorer, MultipleChoiceScorer, SetScorer


class TestScorer(Basetest):
    """
    test for general Scorer
    """

    def testScorer(self):
        """
        test Scorer
        """
        scorer = Scorer()
        scorer.addScore(1, "some check", None)
        scorer.addScore(1, "some check", "some problem")
        scorer.addScore(0.5, "some check", None)
        self.assertEqual(2.5, scorer.maxscore)
        self.assertEqual(1.5, scorer.score)

    def testMultipleChoiceScorer(self):
        """
        test multiple Choice Scorer
        """
        mcs = MultipleChoiceScorer()
        score1 = mcs.evaluate_mctask({1, 2, 5}, {1, 3, 4}, 1, toManyEquals0=True)
        print(score1)

    def testSetScorer(self):
        """
        test Set Scorer
        """
        examples = [
            {"sub": {1, 3, 5, 7, 8}, "expected": {1, 2, 5, 6}, "score": 0},
            {"sub": {1, 2, 3}, "expected": {1, 2, 3}, "score": 1},
            {"sub": {1, 2}, "expected": {1, 2, 3}, "score": 2 / 3},
            {"sub": {1, 2, 3}, "expected": {1, 2}, "score": 1 / 2},
            {"sub": {1, 2, 3}, "expected": {1, 2, 4}, "score": 1 / 3},
        ]
        debug = True
        doTest = True
        for i, example in enumerate(examples):
            setScorer = SetScorer()
            sub = example["sub"]
            expected = example["expected"]
            expectedScore = example["score"]
            score = setScorer.evaluate_set(sub, expected, max_points=1)
            if debug:
                print(
                    f"{i+1}:{sub} expected: {expected} score -> {score} vs expected Score {expectedScore}"
                )
            if doTest:
                self.assertEqual(score, expectedScore)
